/*=============================================================================
Vitalnix User Management Suite
libvxpdb/loader.c
  Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2003 - 2006
  -- License restrictions apply (LGPL v2.1)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.LGPL2"
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX.h>
#include "vitalnix-config.h"
#include "libvxcore/loader.h"
#include "libvxpdb/libvxpdb.h"
#include "libvxutil/defines.h"

enum {
    CONFIG_READ = 1,
    CONFIG_FREE,
};

struct opt {
    char *def_backend;
};

// Functions
static int vxpdb_cleanup(struct vxpdb_state *, int, struct opt *);
static void vxpdb_config(struct opt *, unsigned int);
static void *vxpdb_get_handle(const char *);

static int vdummy_init(struct vxpdb_state *, const void *);
static int vdummy_open(struct vxpdb_state *, long);
static int vdummy_iT(struct vxpdb_state *);
static void vdummy_vT(struct vxpdb_state *);
static long vdummy_modctl(struct vxpdb_state *, long, ...);

static int vdummy_useradd(struct vxpdb_state *, const struct vxpdb_user *);
static int vdummy_usermod(struct vxpdb_state *, const struct vxpdb_user *, const struct vxpdb_user *);
static int vdummy_userdel(struct vxpdb_state *, const struct vxpdb_user *);
static int vdummy_userinfo(struct vxpdb_state *, const struct vxpdb_user *, struct vxpdb_user *, size_t);
static void *vdummy_usertrav_init(struct vxpdb_state *);
static int vdummy_usertrav_walk(struct vxpdb_state *, void *, struct vxpdb_user *);
static void vdummy_usertrav_free(struct vxpdb_state *, void *);

static int vdummy_groupadd(struct vxpdb_state *, const struct vxpdb_group *);
static int vdummy_groupmod(struct vxpdb_state *, const struct vxpdb_group *, const struct vxpdb_group *);
static int vdummy_groupdel(struct vxpdb_state *, const struct vxpdb_group *);
static int vdummy_groupinfo(struct vxpdb_state *, const struct vxpdb_group *, struct vxpdb_group *, size_t);
static void *vdummy_grouptrav_init(struct vxpdb_state *);
static int vdummy_grouptrav_walk(struct vxpdb_state *, void *, struct vxpdb_group *);
static void vdummy_grouptrav_free(struct vxpdb_state *, void *);

//-----------------------------------------------------------------------------
struct vxpdb_state *vxpdb_load(const char *name, void *priv) {
    struct vxpdb_state *new = NULL;
    struct vxpdb_mvtable *vtable;
    struct opt cfg = {};
    int ret;

    vxpdb_config(&cfg, CONFIG_READ);
    if((strcmp(name, "*") == 0 && (name = cfg.def_backend) == NULL)) {
        errno = vxpdb_cleanup(new, EINVAL, &cfg);
        return NULL;
    }

    if(*name == '\0') {
        errno = vxpdb_cleanup(NULL, EINVAL, &cfg);
        return NULL;
    }

    if((new = malloc(sizeof(struct vxpdb_state))) == NULL) {
        errno = vxpdb_cleanup(new, errno, &cfg);
        return NULL;
    }

#ifdef STATIC_BACKENDS
    if((new->handle = vxpdb_get_handle(name)) != NULL) {
        if((vtable = HX_dlsym(new->handle, "THIS_MODULE")) == NULL) {
            errno = pdb_cleanup(new, errno, &cfg);
            return NULL;
        }
    } else if((vtable = vxcore_module_lookup("libvxpdb", name)) == NULL) {
        errno = vxpdb_cleanup(new, errno, &cfg);
        return NULL;
    }
#else
    if((new->handle = vxpdb_get_handle(name)) == NULL ||
     (vtable = HX_dlsym(new->handle, "THIS_MODULE")) == NULL) {
        errno = vxpdb_cleanup(new, errno, &cfg);
        return NULL;
    }
#endif

    new->vtable = vtable;
    vxpdb_fix_vtable(vtable);
    if(vtable->init != NULL && (ret = vtable->init(new, priv)) <= 0) {
        errno = vxpdb_cleanup(new, -ret, &cfg);
        return NULL;
    }

    vxpdb_cleanup(NULL, 0, &cfg);
    return new;
}

void vxpdb_unload(struct vxpdb_state *this) {
    if(this->vtable->deinit != NULL)
        this->vtable->deinit(this);
    vxpdb_cleanup(this, 0, NULL);
    return;
}

/*  vxpdb_fix_vtable
    @m: vtable

*/
void vxpdb_fix_vtable(struct vxpdb_mvtable *m) {
#define SET1(x) if((m->x) == NULL) (m->x) = vdummy_##x;
#define SET2(x, y) if((m->x) == NULL) (m->x) = vdummy_##y;
    SET1(init);
    SET1(open);
    SET2(close, vT);
    SET1(modctl);
    SET2(deinit, vT);
    SET2(lock, iT);
    SET2(unlock, iT);

    SET1(useradd);
    SET1(usermod);
    SET1(userdel);
    if(m->usertrav_init == NULL && m->usertrav_walk == NULL &&
     m->usertrav_free == NULL) {
        m->usertrav_init = vdummy_usertrav_init;
        m->usertrav_walk = vdummy_usertrav_walk;
        m->usertrav_free = vdummy_usertrav_free;
    }
    SET1(userinfo);

    SET1(groupadd);
    SET1(groupmod);
    SET1(groupdel);
    if(m->grouptrav_init == NULL && m->grouptrav_walk == NULL &&
     m->grouptrav_free == NULL) {
        m->grouptrav_init = vdummy_grouptrav_init;
        m->grouptrav_walk = vdummy_grouptrav_walk;
        m->grouptrav_free = vdummy_grouptrav_free;
    }
    SET1(groupinfo);
    return;
}

//-----------------------------------------------------------------------------
static int vxpdb_cleanup(struct vxpdb_state *this, int err, struct opt *cf) {
    if(cf != NULL)
        vxpdb_config(cf, CONFIG_FREE);
    if(this != NULL) {
        if(this->handle != NULL)
            HX_dlclose(this->handle);
        free(this);
    }
    return err;
}

static void vxpdb_config(struct opt *config, unsigned int action) {
    struct HXoption options_table[] = {
        {.ln = "DEFAULT_BACKEND", .type = HXTYPE_STRING,
         .ptr = &config->def_backend},
        HXOPT_TABLEEND,
    };
    if(action == CONFIG_READ)
        HX_shconfig(CONFIG_ETC_VITALNIX "/libvxpdb.conf", options_table);
    else if(action == CONFIG_FREE)
        HX_shconfig_free(options_table);
    return;
}

/*  vxpdb_get_handle
    @filename:  Shared library to open

    Opens @filename or a construction of "drv_" @filename and an extension
    and returns the handle.
*/
static void *vxpdb_get_handle(const char *filename) {
    const char *ext[] = {".so", ".dll", "", NULL}, **extp = ext;
    void *handle;

    // Try plain filename first (as does pdbinfo)
    if((handle = HX_dlopen(filename)) != NULL)
        return handle;

    // If not, prepend prefix, append an extension and try again
    errno = ENOENT;
    while(*extp != NULL) {
        char fn[MAXFNLEN];
        snprintf(fn, MAXFNLEN, "drv_%s%s", filename, *extp);
        if((handle = HX_dlopen(fn)) != NULL) {
            errno = 0;
            break;
        }
        ++extp;
    }
    return handle;
}

//-----------------------------------------------------------------------------
static int vdummy_init(struct vxpdb_state *this, const void *private_data) {
    return 1;
}

static int vdummy_open(struct vxpdb_state *this, long flags) {
    return 1;
}

static void vdummy_vT(struct vxpdb_state *this) {
    return;
}

static int vdummy_iT(struct vxpdb_state *this) {
    return 1;
}

static long vdummy_modctl(struct vxpdb_state *this, long command, ...) {
    switch(command) {
        case PDB_COUNT_USERS:
        case PDB_COUNT_GROUPS:
            return 0;
    }
    return -ENOSYS;
}

static int vdummy_useradd(struct vxpdb_state *this,
 const struct vxpdb_user *user) {
    return -EPERM;
}

static int vdummy_usermod(struct vxpdb_state *this,
 const struct vxpdb_user *mask, const struct vxpdb_user *mod)
{
    return -EPERM;
}

static int vdummy_userdel(struct vxpdb_state *this,
 const struct vxpdb_user *hint)
{
    return -EPERM;
}

static void *vdummy_usertrav_init(struct vxpdb_state *this) {
    return this;
}

static int vdummy_usertrav_walk(struct vxpdb_state *this, void *ptr,
 struct vxpdb_user *result)
{
    return 0;
}

static void vdummy_usertrav_free(struct vxpdb_state *this, void *ptr) {
    return;
}

static int vdummy_userinfo(struct vxpdb_state *this,
 const struct vxpdb_user *mask, struct vxpdb_user *result, size_t size)
{
    return 0;
}

static int vdummy_groupadd(struct vxpdb_state *this,
 const struct vxpdb_group *u) {
    return -EPERM;
}

static int vdummy_groupmod(struct vxpdb_state *this,
 const struct vxpdb_group *srk, const struct vxpdb_group *mod)
{
    return -EPERM;
}

static int vdummy_groupdel(struct vxpdb_state *this,
 const struct vxpdb_group *hint)
{
    return -EPERM;
}

static void *vdummy_grouptrav_init(struct vxpdb_state *this) {
    return this;
}

static int vdummy_grouptrav_walk(struct vxpdb_state *this, void *ptr,
 struct vxpdb_group *result)
{
    return 0;
}

static void vdummy_grouptrav_free(struct vxpdb_state *this, void *ptr) {
    return;
}

static int vdummy_groupinfo(struct vxpdb_state *this,
 const struct vxpdb_group *mask, struct vxpdb_group *result, size_t size)
{
    return 0;
}

//=============================================================================
