/*=============================================================================
Vitalnix User Management Suite
libvxutil/util.c
  Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2003 - 2006
  -- License restrictions apply (LGPL v2.1)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.LGPL2"
=============================================================================*/
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <libHX.h>
#include "libvxutil/defines.h"
#include "libvxutil/libvxutil.h"

// Functions
static char *transform7(const char *, char *, size_t);
static size_t quoted_size(const char *, unsigned int);

// Variables
static const char *quote_match[] = {
    [VXQUOTE_SINGLE] = "'\\",
    [VXQUOTE_DOUBLE] = "\"\\",
    [VXQUOTE_XML]    = "\"&<>",
};

//-----------------------------------------------------------------------------
long vxutil_now_iday(void) {
    return time(NULL) / 86400;
}

int vxutil_only_digits(const char *p) {
    // Returns whether the string in P consists only of digits
    while(*p != '\0')
        if(!isdigit(*p++))
            return 0;
    return 1;
}

char *vxutil_propose_home(char *dest, size_t size, const char *base,
 const char *username, unsigned int level)
{
    if(level == 0)
        snprintf(dest, size, "%s/%s", base, username);
    else if(level == 1 || strlen(username) == 1)
        snprintf(dest, size, "%s/%c/%s", base, *username, username);
    else // level >= 2
        snprintf(dest, size, "%s/%c/%c%c/%s", base, *username,
                 username[0], username[1], username);

    return dest;
}

/*  vxutil_propose_lname
    @surname_8:         The surname (family name), UTF-8 encoded
    @firstname_8:       The first name, UTF-8 encoded
*/
char *vxutil_propose_lname(char *dest, size_t size, const char *surname_8,
 const char *firstname_8)
{
    char surname_7[7], firstname_7[8], *p;

    transform7(firstname_8, firstname_7, sizeof(firstname_7));

    if(surname_8 != NULL) {
        if((p = strchr(surname_8, ' ')) != NULL) // multipart handling
            surname_8 = ++p;
        transform7(surname_8, surname_7, sizeof(surname_7));
        snprintf(dest, size, "%c%s", *firstname_7, surname_7);
    } else {
        snprintf(dest, size, "%s", firstname_7);
    }

    if(strlen(dest) < 2 && size >= 3)
        /* In case the login name is too short, append a "0" (just cosmetics),
        it does not affect anything. */
        strcat(dest, "0");

    return HX_strlower(dest);
#undef MAX_LNAME
}

char *vxutil_quote(const char *src, unsigned int type, char **free_me) {
    char *ret, *ret_wp;

    if(type >= _VXQUOTE_MAX || strpbrk(src, quote_match[type]) == NULL) {
        *free_me = NULL;
        return (char *)src;
    }

    if((ret = ret_wp = *free_me = malloc(quoted_size(src, type) + 1)) == NULL)
        return NULL;

    if(type == VXQUOTE_XML)
        while(*src != '\0') {
            size_t seg_len = strcspn(src, quote_match[type]);
            if(seg_len > 0) {
                memcpy(ret_wp, src, seg_len);
                ret_wp += seg_len;
                src    += seg_len;
                if(*src == '\0')
                    break;
            }
            switch(*src++) {
                case '"':
                    memcpy(ret_wp, "&quot;", 6);
                    ret_wp += 6;
                    break;
                case '&':
                    memcpy(ret_wp, "&amp;", 5);
                    ret_wp += 5;
                    break;
                case '<':
                    memcpy(ret_wp, "&lt;", 4);
                    ret_wp += 4;
                    break;
                case '>':
                    memcpy(ret_wp, "&gt;", 4);
                    ret_wp += 4;
                    break;
            }
            *ret_wp = '\0';
            return ret;
        }

    while(*src != '\0') {
        size_t seg_len = strcspn(src, quote_match[type]);
        if(seg_len > 0) {
            memcpy(ret_wp, src, seg_len);
            ret_wp += seg_len;
            src    += seg_len;
            if(*src == '\0')
                break;
        }
        *ret_wp++ = '\\';
        *ret_wp++ = *src++;
    }

    *ret_wp = '\0';
    return ret;
}

int vxutil_replace_run(const char *fmt, const struct HXoption *map) {
    char cmd[4 * MAXLNLEN];
    HX_strrep5(fmt, map, cmd, sizeof(cmd));
    return system(cmd);
}

char *vxutil_slurp_file(const char *fn) {
    struct stat sb;
    char *dst;
    FILE *fp;

    if((fp = fopen(fn, "rb")) == NULL)
        return NULL;
    if(fstat(fileno(fp), &sb) != 0)
        return NULL;
    if((dst = calloc(1, sb.st_size + 1)) == NULL)
        return NULL;
    if(fread(dst, sb.st_size, 1, fp) < 1) {
        free(dst);
        return NULL;
    }

    fclose(fp);
    return dst;
}

long vxutil_string_iday(const char *s) {
    struct tm td = {};
    int ret = 0;
    time_t r;

    if(strchr(s, '-') != NULL)
        /* ISO-8601 style YYYY-MM-DD, http://www.cl.cam.ac.uk/~mgk25/iso-time.html */
        ret = sscanf(s, "%u-%u-%u", &td.tm_year, &td.tm_mon, &td.tm_mday);
    else if(strchr(s, '.') != NULL)
        // European style DD.MM.YYYY
        ret = sscanf(s, "%u.%u.%u", &td.tm_mday, &td.tm_mon, &td.tm_year);
    else if(strchr(s, '/') != NULL)
        // American style MM/DD/YYYY
        ret = sscanf(s, "%u/%u/%u", &td.tm_mon, &td.tm_mday, &td.tm_year);
    if(ret != 3)
        return -1;

    /* If any of the three are zero, the input date is illegal. If all three
    are zero, the equivalent meaning is "no date given". */
    if(td.tm_mday == 0 && td.tm_mon == 0 && td.tm_year == 0)
        return 0;
    if(td.tm_mday == 0 || td.tm_mon == 0 || td.tm_year == 0)
        return -1;

    if(td.tm_year >= 100 && td.tm_year < 1000) {
        td.tm_year += 1900;
    } else if(td.tm_year < 100) {
        /* Fix two-digit year numbers. The range of two-digit years is always
        [now-50yrs ... now+50yrs]. */
        time_t now_sec = time(NULL);
        struct tm now_tm;
        int bp, nc;

        localtime_r(&now_sec, &now_tm);
        bp = (now_tm.tm_year + 50) % 100;
        nc = now_tm.tm_year - now_tm.tm_year % 100 + ((bp < 50) ? 100 : 0);
        td.tm_year += 1900 + ((td.tm_year > bp) ? (nc - 100) : nc);
    }

    td.tm_year -= 1900;
    --td.tm_mon;

    if((r = mktime(&td)) == -1)
        return -1;
    return (unsigned long)r / 86400;
}

int vxutil_valid_username(const char *n) {
    if(*n == '\0')
        return 0;
    if(!((*n >= 'A' && *n <= 'Z') || (*n >= 'a' && *n <= 'z') || *n == '_'))
        return 0;

    while(*n != '\0') {
        int valid;

        if(*n == '$' && *(n+1) == '\0') // Samba stuffage
            return 1;

        valid = (*n >= 'A' && *n <= 'Z') || (*n >= 'a' && *n <= 'z') ||
                (*n >= '0' && *n <= '9') || *n == '_' || *n == '.' ||
                *n == '-';
        if(!valid)
            return 0;
        ++n;
    }

    return 1;
}

//-----------------------------------------------------------------------------
/*  transform7
    @src:       Source string to transform
    @dest:      Destination buffer
    @dsize:     Size of destination buffer

    Transform an UTF-8 string @src into a 7-bit clean string according to the
    substitution table and put the result into @dest, which is of size @dsize.
    (So it puts in at most @dsize-1 characters.) The result will always be
    '\0'-terminated.
*/
static char *transform7(const char *src, char *dest, size_t dsize) {
    static const struct stab {
        const char *in, *out;
    } subst_tab[] = {
        {"ß", "ss"},
        {"à", "a"}, {"á", "a"}, {"â", "a"}, {"ã", "a"}, {"ä", "ae"}, {"å", "a"}, {"æ", "a"},
        {"è", "e"}, {"é", "e"}, {"ê", "e"}, {"ë", "e"},
        {"ì", "i"}, {"í", "i"}, {"î", "i"}, {"ï", "i"},
        {"ò", "o"}, {"ó", "o"}, {"ô", "o"}, {"õ", "o"}, {"ö", "oe"}, {"ø", "o"},
        {"ù", "u"}, {"ú", "u"}, {"û", "u"}, {"ü", "ue"},
        {"À", "a"}, {"Á", "a"}, {"Â", "a"}, {"Ã", "a"}, {"Ä", "ae"}, {"Å", "a"}, {"Æ", "ae"},
        {"È", "e"}, {"É", "e"}, {"Ê", "e"}, {"Ë", "e"},
        {"Ì", "i"}, {"Í", "i"}, {"Î", "i"}, {"Ï", "i"},
        {"Ò", "o"}, {"Ó", "o"}, {"Ô", "o"}, {"Õ", "o"}, {"Ö", "oe"}, {"Ø", "o"},
        {"Ù", "u"}, {"Ú", "u"}, {"Û", "u"}, {"Ü", "ue"},
        {"ç", "c"}, {"Ç", "c"},
        {"Ñ", "n"}, {"ñ", "n"},
        {NULL},
    };

    char *od = dest;
    --dsize;
    while(dsize && *src != '\0') {
        if((*src >= 'A' && *src <= 'Z') || (*src >= 'a' && *src <= 'z')) {
            *dest++ = *src++;
            --dsize;
        } else if(*(unsigned char *)src >= 128) {
            const struct stab *sp = subst_tab;
            int ok = 0;
            while(sp->in != NULL) {
                if(strstr(src, sp->in) == src) {
                    size_t l = strlen(sp->out);
                    if(dsize >= l) {
                        strncpy(dest, sp->out, l);
                        src   += strlen(sp->in);
                        dest  += l;
                        dsize -= l;
                        ++ok;
                        break;
                    } else {
                        /* Do not let "abcdeöa" through, because if there was
                        too few space, the "ö" gets skipped but dsize is still
                        big enough to let the "a" in. */
                        dsize = 0;
                    }
                    break;
                }
                ++sp;
            }
            // Just skip things we can not transform
            if(!ok) ++src;
        } else {
            ++src;
        }
    }
    *dest = '\0';
    return od;
}

/*  quoted_size
    @s:         string to analyze
    @type:      non-zero if double quoted

    Returns the size of the string @s when quoted.
*/
static size_t quoted_size(const char *s, unsigned int type) {
    const char *p = s;
    size_t n = strlen(s);

    if(type == VXQUOTE_XML) {
        while((p = strpbrk(p, quote_match[type])) != NULL) {
            switch(*p) {
                // minus 1 because the original character will disappear
                case '"':
                    n += 6 - 1; // &quot;
                    break;
                case '&':
                    n += 5 - 1; // &amp;
                case '<':
                case '>':
                    n += 4 - 1; // &lt; and &gt;
                    break;
            }
            ++p;
        }
        return n;
    }

    while((p = strpbrk(p, quote_match[type])) != NULL) {
        ++n;
        ++p;
    }

    return n;
}

//=============================================================================
